package mct.admin;


/**
 * The class forwards internal events in the Mobile Code Manager
 * to all registered listeners.
 * Although patterned after AWTEventListener, the central idea
 * in this implementation is to isolate the Mobile Code Manager (MCM)
 * from the listeners : in contrast to the AWT, the MCM's thread 
 * shall not directly invoke the listeners' event handlers 
 * because there are no guarantees that the listeners are written 
 * with the integrity of the MCM in mind. By having the MCM simply 
 * queue events and having a thread within this class, we allow the 
 * MCM to continue on with its own work.
 *
 * The overhead of event-monitoring must be minimized through
 * judicious declaration of events, and perhaps later, some
 * enable/disable methods
 *
 * TBD : I want to use reflection for this class.
 * Q : What happens if a netlet implements >1 listener.  Will it be
 *	registered for >1 type of event ?
 */

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import mct.users.*;
public class MCTEventMulticaster implements MCEventListener, CallbackEventListener, MigratableEventListener, MCTGenericEventListenerInterface, Runnable {

	//**********
	//* Fields *
	//**********

	public static String MCEVENT = "MCEvent";
	public static String CALLBACK_EVENT = "CallbackEvent";
	public static String MIGRATABLE_EVENT = "MigratableEvent";
	public static String GENERIC_EVENT = "GenericEvent";
	private Vector eventQueue; // FIFO queue of events internal to MCM
	private Hashtable listeners; // Table of listeners registered for 
	// different types of events
	// The key is the type of event
	// The entry is a vector of registered listeners

	Thread dispatcher = null; // The thread that empties the event queue













/** 
   * Constructs the Event Multicaster for the Mobile Code Toolkit.
   * 
   */
public MCTEventMulticaster() {
	eventQueue = new Vector();
	listeners = new Hashtable();
	Class interfaces[] = this.getClass().getInterfaces();
	listeners.put(MCEVENT, new Vector());
	listeners.put(CALLBACK_EVENT, new Vector());
	listeners.put(MIGRATABLE_EVENT, new Vector());
	listeners.put(GENERIC_EVENT, new Vector());
	dispatcher = new Thread(this);
	dispatcher.start();
}
public void addCallbackEventListener(CallbackEventListener listener) {
	((Vector) listeners.get(CALLBACK_EVENT)).addElement(listener);
}
private void addEvent(String listenerClass, String method, MCTEvent event) {
	eventQueue.addElement(new QueueEntry(listenerClass, method, event));
	dispatcher.resume();
}
/**
 * This method was created by a SmartGuide.
 * @param listener mct.users.MCTGenericEventListener
 */
public void addGenericEventListener(MCTGenericEventListener listener) {
	((Vector) listeners.get(GENERIC_EVENT)).addElement(listener);
	return;
}
public void addMCEventListener(MCEventListener listener) {
	((Vector) listeners.get(MCEVENT)).addElement(listener);
}
public void addMigratableEventListener(MigratableEventListener listener) {
	((Vector) listeners.get(MIGRATABLE_EVENT)).addElement(listener);
}
public void addMobileCode(MCEvent event) {
	addEvent(MCEVENT, "addMobileCode", event);
}
public void migrated(MigratableEvent event) {
	addEvent(MIGRATABLE_EVENT, "migrated", event);
}
public void netletInitialized(CallbackEvent event) {
	addEvent(CALLBACK_EVENT, "netletInitialized", event);
}
public void netletResumed(CallbackEvent event) {
	addEvent(CALLBACK_EVENT, "netletResumed", event);
}
public void netletStarted(CallbackEvent event) {
	addEvent(CALLBACK_EVENT, "netletStarted", event);
}
public void netletStopped(CallbackEvent event) {
	addEvent(CALLBACK_EVENT, "netletStopped", event);
}
public void netletSuspended(CallbackEvent event) {
	addEvent(CALLBACK_EVENT, "netletSuspended", event);
}
/**
 * This method was created by a SmartGuide.
 * @param event mct.users.MCTGenericEvent
 */
public void notify(MCTGenericEvent event) {
	addEvent(GENERIC_EVENT, "generic event", event);
	return;
}
public void removeCallbackEventListener(CallbackEventListener listener) {
	((Vector) listeners.get(CALLBACK_EVENT)).removeElement(listener);
}
/**
 * This method was created by a SmartGuide.
 * @param listener mct.users.MCTGenericEventListener
 */
public void removeGenericEventListener(MCTGenericEventListener listener) {
	((Vector) listeners.get(GENERIC_EVENT)).removeElement(listener);
	return;
}
public void removeMCEventListener(MCEventListener listener) {
	((Vector) listeners.get(MCEVENT)).removeElement(listener);
}
/**
 * This method was created by a SmartGuide.
 * @param mc mct.users.MobileCodeObject
 */
public synchronized void removeMCTGenericEventListener(MobileCodeObject mc) {
	Vector registeredListeners = (Vector) listeners.get(GENERIC_EVENT);
	Receiver receiver = new Receiver(mc, null);
	int index;
	if (registeredListeners != null) {
		// There are listeners registered for this class of events
		// We have to try to remove listeners multiple times because the 
		// mobile code may have registered for several types of event.
		while ((index = registeredListeners.indexOf(receiver)) >= 0)
			registeredListeners.removeElementAt(index);
	}
	return;
}
public void removeMigratableEventListener(MigratableEventListener listener) {
	((Vector) listeners.get(MIGRATABLE_EVENT)).removeElement(listener);
}
public void removeMobileCode(MCEvent event) {
	addEvent(MCEVENT, "removeMobileCode", event);
}
public void run() {
	for (;;) {
		try {
			// Get the next event, if available (see catch below)
			Object next;
			synchronized (eventQueue) {
				// Don't want an event inserted between the fetch and the removal 
				// of the first element
				next = eventQueue.elementAt(0);
				eventQueue.removeElementAt(0);
			}
			QueueEntry event = (QueueEntry) next;

			//* Class c = Class.forName (event.listenerClass);
			//* Method m = c.getMethod (event.method, parameters);
			//* m.invoke (c, event);

			// Notify all registered listeners of this event
			Vector registeredListeners = (Vector) listeners.get(event.listenerClass);
			if (registeredListeners != null) {
				// There are listeners registered for this class of events
				for (Enumeration e = registeredListeners.elements(); e.hasMoreElements();) {
					Object listener = e.nextElement();
					if (event.listenerClass == MCEVENT) {
						//  System.out.println ("GOT MC Event");
						if (event.method == "addMobileCode") {
							((MCEventListener) listener).addMobileCode((MCEvent) event.arg);
						} else
							if (event.method == "removeMobileCode") {
								((MCEventListener) listener).removeMobileCode((MCEvent) event.arg);
							} else {
								System.err.println("");
							}
					} else
						if (event.listenerClass == MIGRATABLE_EVENT) {
							// System.out.println ("GOT Migratable Event");
							if (event.method == "migrated") {
								((MigratableEventListener) listener).migrated((MigratableEvent) event.arg);
							} else {
								System.err.println("");
							}
						} else
							if (event.listenerClass == CALLBACK_EVENT) {
								// System.out.println ("GOT Callback Event");
								if (event.method == "netletInitialized") {
									((CallbackEventListener) listener).netletInitialized((CallbackEvent) event.arg);
								} else
									if (event.method == "netletStarted") {
										((CallbackEventListener) listener).netletStarted((CallbackEvent) event.arg);
									} else
										if (event.method == "netletStopped") {
											((CallbackEventListener) listener).netletStopped((CallbackEvent) event.arg);
										} else
											if (event.method == "netletSuspended") {
												((CallbackEventListener) listener).netletSuspended((CallbackEvent) event.arg);
											} else
												if (event.method == "netletResumed") {
													((CallbackEventListener) listener).netletResumed((CallbackEvent) event.arg);
												}
							}
							// Handle the generic events using reflection. 
							else
								try {
								((MCTGenericEventListener) listener).process((MCTGenericEvent) event.arg);
							} catch (java.lang.IllegalAccessException ex) {
					} catch (java.lang.reflect.InvocationTargetException ex) {
					}
				}
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// No element is available. Wait.
			dispatcher.suspend();
		}
	}
}
}