package mct.blackboard;


/**
 *@(#)MsgBlackBoard.java    
 */
/*
 *  This is a stationary netlet that implements a blackboard whereby
 *  messages are posted according to a key, and netlets
 *  may either directly query the messages or register to 
 *  be notified of any postings as they occur.
 * 
 *  In this version, blackboard operations are implemented
 *  with Messages sent via the Facilitator (ie. MessagingAccess)
 *  An alternative implementation could use its own direct
 *  messaging protocol or RMI.
 * 
 *  In this version, one message is posted per key.
 *  In later versions, messages will be stored in groups,
 *  with buffering (>1 message per key) and also different
 *  organizations (eg. >1 key per possible per message)
 *
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import mct.admin.MobileCodeStub;
import mct.users.SuperNetlet;
import mct.users.MessagingAccess;
import mct.mediator.Message;
public class MsgBlackBoard extends SuperNetlet implements BlackBoard, MessagingAccess {
	private Hashtable registrants; // Lists of netlets registered for a key
	private Hashtable board; // Posted messages, stored ONE per key
	private List list = null; // For debugging, visualization
	private int maxTries = 1; // Default value for tries
	private int timer = 1000; // Default value for timeout




/** 
  * Constructs an empty blackboard
  */
public MsgBlackBoard() {
	registrants = new Hashtable();
	board = new Hashtable();
	list = new List();
}
/**
  * Cancel the registration of the mobileCode for the given key
  *
  * @param	mcID	Identifier of mobileCode.
  *			Use getIdentifier.getID() for messaging
  * @param	key	"Type" of postings of interest
  * @returns	True if cancellation is successful
  */
public synchronized boolean deRegister(String mcID, Object key) {
	Vector ids;
	if (registrants.containsKey(key)) {
		ids = (Vector) registrants.get(key);
		ids.removeElement(mcID);
		registrants.put(key, ids);
	}
	return true;
}
/**
  * Retrieve any posting on the blackboard under the given key
  * @param	mcID	Identifier of mobileCode to which posting should be sent
  * @param	key	Type of posting to be found
  * @returns	A message, if present
  *		Null, if no posting is present for the key
  */
public synchronized Object enquire(String mcID, Object key) {
	Object contents = board.get(key);
	if (contents == null)
		contents = "<UNKNOWN>";
	if (contents != null) {
		EnquireReplyMessage reply = new EnquireReplyMessage(getID(), mcID, key, contents);
		int tries = 0;
		while (!sendAsync(reply) && tries < maxTries) {
			tries++;
			mct.Console.application(this, "unable to send reply to enquire to " + mcID);
			try {
				Thread.sleep(timer);
			} catch (InterruptedException ie) {
			}
		}
	}
	return contents;
}
/**
  * Returns the status of the blackboard
  * @returns 	True if the blackboard is available
  */
public boolean isAvailable() {
	return true;
}
/**
  * 
  */
public boolean onInit(Properties props) {
	if (props != null) {
		String parameter = (String) props.getProperty("blackboard.display.enable");
		if (parameter == null || parameter.equals("true"))
			setupDisplay();
		else
			list = null;
		parameter = (String) props.getProperty("blackboard.messaging.tries");
		if (parameter == null)
			maxTries = 1;
		else
			maxTries = (new Integer(parameter)).intValue();
		parameter = (String) props.getProperty("blackboard.messaging.timeout");
		if (parameter == null)
			timer = 1000;
		else
			timer = (new Integer(parameter)).intValue();
	} else {
		list = null;
		maxTries = 1;
		timer = 1000;
	}
	return true;
}
/**
  * Callback method accepts incoming message from users of the
  * blackboard to either register, post or enquire
  *
  * @see mct.users.MessagingAccess
  */
public void onStart() {
	Message msg = null;
	for (;;) {
		try {
			msg = receive();
			while (msg == null) {
				try {
					Thread.sleep(timer);
					System.out.println(getID()+" sleeping...");
				} catch (java.lang.InterruptedException ex) {
				};
				msg = receive();
			};
			BlackBoardMessage bbmsg = (BlackBoardMessage) msg;
			if (bbmsg instanceof RegisterMessage) {
				RegisterMessage rmsg = (RegisterMessage) bbmsg;
				this.register(rmsg.getFrom(), rmsg.getKey());
			} else
				if (bbmsg instanceof DeRegisterMessage) {
					DeRegisterMessage dmsg = (DeRegisterMessage) bbmsg;
					this.deRegister(dmsg.getFrom(), dmsg.getKey());
				} else
					if (bbmsg instanceof PostMessage) {
						PostMessage pmsg = (PostMessage) bbmsg;
						this.post(pmsg.getKey(), pmsg.getContents());
					} else
						if (bbmsg instanceof EnquireMessage) {
							EnquireMessage emsg = (EnquireMessage) bbmsg;
							this.enquire(emsg.getFrom(), emsg.getKey());
						} else {
							mct.Console.application(this, " received unsupported message type " + bbmsg.toString());
						}
		} catch (ClassCastException ce) {
			mct.Console.application(this, " received unknown message format " + msg);
			ce.printStackTrace();
		} catch (NullPointerException ex) {
			mct.Console.application(this, " null pointer error ");
		}
	}
}
/**
  * Post a message on the blackboard, under the given key
  *
  * @param	key	Type of posting 
  * @param	message	The object posted on the blackboard
  * @returns	True if posting and associated notifications
  *	        are completed successfully
  */
public synchronized boolean post(Object key, Object message) {
	board.put(key, message);
	Vector ids = (Vector) registrants.get(key);
	if (ids != null) {
		for (Enumeration e = ids.elements(); e.hasMoreElements();) {
			NotifyMessage update = new NotifyMessage(getID(), (String) e.nextElement(), key, message);
			int tries = 0;
			while (!sendAsync(update) && tries < maxTries) {
				tries++;
				try {
					Thread.sleep(timer);
					mct.Console.message(null, getID()+" sleeping while notifying...");
				} catch (InterruptedException ie) {
				};
				mct.Console.application(null, getID() + " failed to notify " + update.getTo() + ", try in " + timer + " msec.");
				// return false;
			}
			if (tries == maxTries) {
				mct.Console.application(null, getID() + " deregistering " + update.getTo());
				deRegister(update.getTo(), key);
			}
		}
	}
	if (list != null)
		list.addItem("Posting " + (String) key + " with " + ((Integer) message).intValue());
	return true;
}
/**
  * Register the mobileCode to receive notifications of
  * any posting to the blackboard under the given key
  *
  * @param	mcID	Identifier of mobileCode.
  *			Use getIdentifier.getID() for messaging
  * @param	key	"Type" of postings of interest
  * @returns	True if registration is successful
  */
public synchronized boolean register(String mcID, Object key) {
	Vector ids;
	if (registrants.containsKey(key)) {
		ids = (Vector) registrants.get(key);
	} else {
		ids = new Vector();
	}
	ids.addElement(mcID);
	registrants.put(key, ids);
	return true;
}
/**
 * This method was created by a SmartGuide.
 */
private void setupDisplay() {
	// For debugging, a window is popped up to show up all postings
	Frame frame = new Frame("Message BlackBoard");
	list = new List();
	frame.add(list);
	frame.pack();
	frame.validate();
	frame.show();
	return;
}
}