package mct.management;


import mct.admin.*;
import mct.users.*;
import java.rmi.Naming;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.RemoteException;
/**
 * This type allows for remote management of mobile code.
 * By adding this virtual managed component to the mobile code 
 * manager, mobile code can be started, stopped, suspended or 
 * resumed by remote action using RMI.
 */
public class RemoteManager extends SuperVirtualManagedComponent implements ManagementInterface {
	public boolean available = false;


/**
 * This method creates a new manager
 */
public RemoteManager() throws RemoteException {
	super();
	UnicastRemoteObject.exportObject(this);
}
/**
 * This method creates a new remote manager
 * @param name java.lang.String
 */
public RemoteManager(String name) throws RemoteException {
	super(name);
	UnicastRemoteObject.exportObject(this);
}
/**
 * This method allows a piece of mobile code to be
 * destroyed.
 * @return boolean
 * @param id mct.users.MCIdentifier
 */
public boolean destroy(MCIdentifier id) throws RemoteException {
	try {
		MobileCodeManager mcm = (MobileCodeManager) getMobileCodeManager();
		if (mcm == null)
			return false;
		MobileCodeObject mco = mcm.getMobileCodeHandle(id.getID());
		if (!(mco instanceof MobileCodeObject))
			return false;
		mco.resetEventNumber();
		mcm.destroy(mco);
		return true;
	} catch (Throwable ex) {
		if (ex instanceof RemoteException)
			throw (RemoteException) ex;
		else
			throw new RemoteException(ex.toString());
	}
}
/**
 * This method was created in VisualAge.
 */
public void finalize() {
	available = false;
	try {
		Naming.unbind(getID());
	} catch (Exception e) {
		mct.Console.error(this, "cannot unbind "+getID()+" in RMI registry\nReason: " + e);
	}
}
/**
 * Bind the object in the RMI registry locally. Maybe we'll extend 
 * this to add the entry in several registries for fault tolerance.
 * @param properties java.util.Properties
 * @return boolean
 */
public boolean onInit(java.util.Properties properties) {
	try {
		Naming.rebind(getID(), this);
		available = true;
	} catch (Exception e) {
		mct.Console.error(this, "cannot rebind "+getID()+" in RMI registry.\nReason: "+e);
		return false;
	}
	return true;
}
/**
 * This method tells the user that the remote manager is
 * up and running. Of course, the message may never appear
 * if the display options aren't set up.
 */
public void onStart() {
	mct.Console.message(null, getID()+" is installed");
	return;
}
/**
 * This method resumes a piece of mobile code. The value true is returned
 * if the mobile code can be resumed, false otherwise.
 * @return boolean
 * @param id mct.users.MCIdentifier
 */
public boolean resume(MCIdentifier id) throws RemoteException {
	try {
		MobileCodeManager mcm = (MobileCodeManager) getMobileCodeManager();
		if (mcm == null)
			return false;
		MobileCodeObject mco = mcm.getMobileCodeHandle(id.getID());
		if (!(mco instanceof MobileCodeObject))
			return false;
		mcm.resume(mco);
		return true;
	} catch (Throwable ex) {
		if (ex instanceof RemoteException)
			throw (RemoteException) ex;
		else
			throw new RemoteException(ex.toString());
	}
}
/**
 * This method stops a piece of mobile code. The value true is returned
 * if the mobile code can be stopped, false otherwise.
 * @return boolean
 * @param id mct.users.MCIdentifier
 */
public boolean start(MCIdentifier id) throws RemoteException {
	try {
		MobileCodeManager mcm = (MobileCodeManager) getMobileCodeManager();
		if (mcm == null)
			return false;
		MobileCodeObject mco = mcm.getMobileCodeHandle(id.getID());
		if (!(mco instanceof MobileCodeObject))
			return false;
		mco.nextEvent();
		mcm.start(mco);
		return true;
	} catch (Throwable ex) {
		if (ex instanceof RemoteException)
			throw (RemoteException) ex;
		else
			throw new RemoteException(ex.toString());
	}
}
/**
 * This method stops a piece of mobile code. The value true is returned
 * if the mobile code can be stopped, false otherwise.
 * @return boolean
 * @param id mct.users.MCIdentifier
 */
public boolean stop(MCIdentifier id) throws RemoteException {
	try {
		MobileCodeManager mcm = (MobileCodeManager) getMobileCodeManager();
		if (mcm == null)
			return false;
		MobileCodeObject mco = mcm.getMobileCodeHandle(id.getID());
		if (!(mco instanceof MobileCodeObject))
			return false;
		mco.nextEvent();
		mcm.stop(mco);
	} catch (Throwable ex) {
		if (ex instanceof RemoteException)
			throw (RemoteException) ex;
		else
			throw new RemoteException(ex.toString());
	}
	return true;
}
/**
 * This method suspends a piece of mobile code. The value true is returned
 * if the mobile code can be suspended, false otherwise.
 * @return boolean
 * @param id mct.users.MCIdentifier
 */
public boolean suspend(MCIdentifier id) throws RemoteException {
	try {
		MobileCodeManager mcm = (MobileCodeManager) getMobileCodeManager();
		if (mcm == null)
			return false;
		MobileCodeObject mco = mcm.getMobileCodeHandle(id.getID());
		if (!(mco instanceof MobileCodeObject))
			return false;
		mcm.suspend(mco);
		return true;
	} catch (Throwable ex) {
		if (ex instanceof RemoteException)
			throw (RemoteException) ex;
		else
			throw new RemoteException(ex.toString());
	}
}
}