package mct.mediator;


// TBD 1 

/**
 *@(#)CommunicationFacilitator.java    	Cheryl Schramm May 1998
 *
 * The Facilitator is a MobileCodeDaemon extension that provides
 * a messaging service to the mobile code currently resident at this
 * "place".  Minimally, the Facilitator requires a local directory
 * of mobile code to provide local messaging.  For remote messaging,
 * a regional directory (called a mediator in the code) is also
 * required.
 *
 * At the heart of the Facilitator are named-mailboxes : synchronization
 * objects whereby a message is passed from a sender to a receiver,
 * the receiver waiting if necessary until the sender's deposit unblocks
 * the receiver.  
 *
 * For sending a message : if the destination is found in the local
 * directory, the message is deposited directly into the destination's
 * local inMbx; otherwise, the remote directory is queried for the
 * current location of the destination, and the message is forwarded
 * to this place.  In either case, the sender then waits on its local
 * ackMbx for an acknowledgement of receipt.
 *
 * For receiving a message : the mobile code retrieves a message
 * from its local inMbx, waiting if necessary.  An acknowledgement
 * message is sent to the originator, using the same transport as
 * for sending a message, above, with the message being deposited
 * in the ackMbx. 
 *
 * For receiving forwarded messages : a separate thread listens
 * on a port for messages that are being forwarded from other places.
 * The incoming message is put on the destination's inMbx, unless
 * it is an Acknowledgment message in which case it is put on the ackMbx.
 *
 */

import java.net.*;
import java.net.UnknownHostException;
import java.util.*;
import java.io.*;
import java.rmi.*;
import mct.admin.*;
import mct.users.*;
public class CommunicationFacilitator extends MobileCodeObject implements MobileCode, MCDInterface, CommunicationFacilitatorInterface, Runnable {

	// Visual Age does not support inner classes.  So...
	// These variables have default visibility, giving access to
	// related classes in the same package.


	//* Startup properties

	String MCDirectory = null; // Parameter : Classname of the "residents" directory
	Properties MCDirProperties = null; // Parameter : Property file for "residents" directory
	String mediatorListener = null; // Parameter : Classname of regional directory
	InetAddress mediatorIP = null; // Parameter : Location of regional directory
	int mediatorPort = -1; // Parameter : Listening port of regional directory
	int facPort = -1; // Parameter : Listening port of this facilitator

	//* References to other objects

	private MobileCodeManager mcm = null;
	private ServerSocket listener = null; // Listening socket for incoming messages
	private Thread listenerThread = null; // Separate thread for incoming messages

	//* Major messaging data structures

	MobileCodeDirectory residents = null; // Directory of all local mobile codes
	Hashtable ackMbx = null; // Acks for local netlets
	Hashtable inMbx = null; // Incoming messages for local netlets

	private RemoteMCDirectory mediator; // For outgoing messages (forwarding)
	private String mediatorName = "Mediator";
	private Vector mediators = null; // Vector of MessagingProtocols, for
	// connections to multiple mediators 
	private boolean available; // True when naming services are setup











/** 
   * Constructs the major data structures in the Facilitator and
   * starts up the listening thread
   * WARNING : No MobileCodeStub methods can be called in the constructor
   */
public CommunicationFacilitator() {
	available = false;

	// The facilitator's identifier must match the search in MobileCodeManager.send()&receive()
	ackMbx = new Hashtable();
	inMbx = new Hashtable();
	mediators = new Vector();
}
/**
   * Delivers a message to a local netlet via its threaded mailbox.
   * Notifies the netlet that the message has arrived, if supported.
   */
protected boolean deliverMessage(MobileCodeObject dest, Message message) {
	if (dest instanceof MessagingAccess) {
		//MobileCodeObject mCode = mcm.getMobileCodeHandle(dest.id);
		if (dest == null)
			return false;
		sendMbx(inMbx, dest.getID(), message);
		return true;
	} else {
		mct.Console.warning(null, "deliverMessage failed to " + dest + " because noMessagingAccess");
		return false;
	}
}
public void finalize() {
	try {
		listener.close();
	} catch (IOException e) {
		mct.Console.error(this, "Unable to close listener");
		e.printStackTrace();
	}
}
/**
   * Routes the outgoing message to a remote netlet via the mediator
   * It is synchronized because the "mediator" connection is shared among
   * all sending netlets
   * Returns false when no mediator is available, local messaging only
   *
   * This implementation assumes that mediator provides the message
   * delivery service. An alternative is to simply query the mediator
   * to retrieve the dest's location, and then deliver it straight
   * to that location.  Protected allows you to override.
   */

protected boolean forwardMessage(Message message) {
	String location = null;
	MessagingProtocol dest = null;
	if (mediator != null) {
		try {
			MobileCodeRecord mcr = mediator.getMobileCodeRecordR(message.getTo());
			if (mcr == null) {
				/* There are two reasons why an empty MobileCodeRecord will be returned
				   *   1) There is no such netlet
				   *   2) That netlet is in transit (migration)
				   */
				for (int i = 0; i < 5 && (mcr == null); i++) {
					// yield ();	// yield() is ideal but does not seem to do the job
					try {
						Thread.sleep(2000);
					} catch (InterruptedException ie) {
					}
					mcr = mediator.getMobileCodeRecordR(message.getTo());
				}
				if (mcr == null) {
					mct.Console.warning(null, getID() + " " + message.getTo() + " is not registered in mediator");
					return false;
				}
			}
			location = mcr.location;

			// Parse out the hostname and the facport
			int index = location.indexOf(':');
			String host = location.substring(0, index);
			index = location.lastIndexOf(':');
			int facPort = Integer.parseInt(location.substring(index + 1));
			dest = new MessagingProtocol(InetAddress.getByName(host), facPort);
			dest.send(message);
			byte ack = dest.waitACK();
			dest.close();
			if (ack == MessagingProtocol.NO_ERROR) {
				return true;
			}
		} catch (RemoteException e) {
			mct.Console.error(null, getID() + ": forwarding exception: " + e.getMessage());
		} catch (java.net.UnknownHostException ue) {
			mct.Console.error(null, getID() + ": forwarding exception: " + ue.getMessage());
		} catch (IOException ioe) {
			mct.Console.error(null, getID() + ": unable to make connection to " + location + ". Message is not sent.");
			if (dest != null)
				dest.close();
		}
	} else {
		mct.Console.error(null, getID() + ": no mediator available. Local messaging only.");
		// Could also try to lookup ... subscribeToMediator
	}
	return false;
}
/**
   * @see	mct.users.MCDExtension
   */
public MCTEventMulticaster getEventDispatcher() throws SecurityException {
	MobileCodeStub stub = getStub();
	if (stub != null) {
		return stub.getEventDispatcher();
	}
	return null;
}
/**
   * Creates a mailbox for a destination if necessary. It is the mailbox's
   * responsibility to invoke a callback on the object.
   * @return mct.mediator.MailboxInterface
   * @param	mbxes	Table of mailboxes, indexed by messengers' names
   * @param	dest	Name of destination
   * @param	mobileCodeObject Used to create a threaded mailbox if necessary
   */
private MailboxInterface getMbx(Hashtable mbxes, Object dest, MobileCodeObject mobileCodeObject) {
	if (!mbxes.containsKey(dest)) {
		if (mbxes == inMbx) {
			Receiver rcvr;
			if (mobileCodeObject != null)
				rcvr = new Receiver(mobileCodeObject, "onMessage", new Class[0]);
			else
				rcvr = null;
			mbxes.put(dest, new ThreadedMailbox(rcvr));
		} else
			mbxes.put(dest, new Mailbox());
	}
	return (MailboxInterface) mbxes.get(dest);
}
/**
   * Returns the IP address of the remote Mobile Code Directory, used
   * for remote messaging
   */
public InetAddress getMediatorIP() {
	return mediatorIP;
}
/**
   * Returns the listening port of the remote Mobile Code Directory, used
   * for remote messaging
   */
public int getMediatorPort() {
	return mediatorPort;
}
/**
   * @see	mct.users.MCDExtension
   */
public MobileCodeContainer getMobileCodeContainer() {
	MobileCodeStub stub = getStub();
	if (stub == null)
		return null;
	return stub.getMobileCodeContainer();
}
/**
   * @see	mct.users.MCDExtension
   */
public Object getMobileCodeManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getMobileCodeManager();
	return null;
}
/** 
   * Returns the port on which the Facilitator is listening for messages
   * forwarded from other places
   */
public int getPort() {
	return facPort;
}
/**
   * @see	mct.users.MCDExtension
   */
public StorageManager getStorageManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getStorageManager();
	return null;
}
/**
 * This method was created by a SmartGuide.
 * @return java.lang.String
 */
public String getType() {
	return "Facilitator";
}
/**
 * This method was created by a SmartGuide.
 * @return boolean
 * @param dest java.lang.String
 */
public boolean hasMessages(String dest) {
	if (!available)
		return false;
	MailboxInterface mbx;
	if (!inMbx.containsKey(dest))
		return false;
	else
		mbx = (MailboxInterface) inMbx.get(dest);
	return mbx.hasMessages();
}
/**
   * Returns the current availability of the facilitator
   * @return	False if facilitator is down due to init errors
   *		True if facilitator is able to forward messages
   */
public boolean isAvailable() {
	return available;
}
/**
  * Use the properties to initialize local variables
  *
  * @param properties java.util.Properties
  * @returns	True if no errors in the properties were encountered
  */
public boolean onInit(java.util.Properties properties) {
	String mediatorListenerProperties;
	available = false;
	// Set variables defined as properties
	if (properties == null)
		return false;
	String s;
	try {
		if ((MCDirectory = properties.getProperty("fac.directory.mobilecode")) == null)
			return false;
		String directoryProperties = properties.getProperty("fac.directory.properties");
		if (directoryProperties != null) {
			try {
				FileInputStream fis = new FileInputStream(directoryProperties);
				MCDirProperties = new Properties();
				MCDirProperties.load(fis);
				fis.close();
			} catch (IOException ex) {
				mct.Console.warning(this, "could not load directory properties.\n" + ex);
				MCDirProperties = null;
			}
		}
		facPort = Integer.parseInt(properties.getProperty("fac.port", "6666"));
		mediatorName = properties.getProperty("fac.mediator.name", mediatorName);
		mediatorPort = Integer.parseInt(properties.getProperty("fac.mediator.port", "8888"));
		if ((s = properties.getProperty("fac.mediator.ip")) != null) {
			mediatorIP = InetAddress.getByName(s);
		} else {
			return false;
		}
		if ((mediatorListener = properties.getProperty("fac.mediator.listener")) == null)
			return false;
		mediatorListenerProperties = properties.getProperty("fac.mediator.listener.properties");
	} catch (UnknownHostException ue) { // TBD : Have separate error-handling messages
		return false;
	} catch (ClassCastException ce) {
		return false;
	} catch (Exception e) {
		return false;
	}

	// Get the handle to the Mobile Code Manager, to facilitate 
	// later enquiries about netlets

	if ((mcm = (MobileCodeManager) getMobileCodeManager()) == null) {
		return false;
	}

	// Install the local MobileCodeDirectory, called residents

	// mcm.instantiate (MCDirectory, MCDirProp);
	// TBD 1 : Calling instantiate as above assumes that the
	//	class can be loaded from the CLASSPATH.
	//  To handle the general case of loading from a URL,
	//	your first load the jarTable
	// Ultimately, if we let MCDirectory have properties :
	boolean ok = mcm.instantiateWithProperties(MCDirectory, MCDirProperties);
	if (ok) {
		MobileCodeObject temp;
		while ((temp = mcm.getMobileCodeHandle(MCDirectory)) == null) {
			Thread.yield();
		}
		if (temp instanceof MobileCodeDirectory) {
			residents = (MobileCodeDirectory) temp;
			try {
				residents.setTitle(temp.getClass().getName() + " " + InetAddress.getLocalHost().toString() + ":" + properties.getProperty("fac.port"));
			} catch (java.net.UnknownHostException ex) {
				residents.setTitle(temp.getClass().getName() + " <UNKNOWN HOST>:" + properties.getProperty("fac.port"));
			}
			mct.Console.message(this, "Mobile code directory is active");
		} else {
			mct.Console.warning(this, "Could not load MobileCodeDirectory. No message service available");
			residents = null;
			return false;
		}
		while (!residents.isAvailable()) {
			try {
				Thread.sleep(150);
			} catch (java.lang.InterruptedException ex) {
			}
		}

		// Register with a local mediator for referral services

		mediator = subscribeToMediator(mediatorListener, mediatorListenerProperties, mediatorIP, mediatorPort);
		//  For each mediator defined in the properties file...
		//  mediators.addElement = subscribeToMediator (mediatorIP, mediatorPort);

		//* Set up socket for listening for incoming messages
		try {
			listener = new ServerSocket(facPort);
		} catch (IOException e) {
			mct.Console.error(this, "Unable to open listening socket at port " + facPort);
			return false;
		}
		listenerThread = new Thread(this);
		listenerThread.start();
		return true;
	} else
		return false;
}
/**
 * This method was created by a SmartGuide.
 * @return mct.mediator.Message
 * @param id java.lang.String
 */
public Message peek(String dest) {
	if (!available)
		return null;
	MailboxInterface mbx;
	if (!inMbx.containsKey(dest))
		return null;
	else
		mbx = (MailboxInterface) inMbx.get(dest);
	if (mbx instanceof MailboxInterface)
		return (Message)mbx.peek();
	return null;
}
/**
 * The thread's body : Read and decipher the message sent on the given socket
 *	If the message is a Request, deposit the message in the destination's mailbox
 *	If the message is a Response, deposit the message in the destination's ack mailbox,
 *		as an end-to-end acknowledgement to a message previously sent.
 *
 * NOTE : There are two levels of acknowledgements in the Facilitator :
 *	a) application level : acknowledgement between netlets (automatically
 *	   	sent by Facilitator) signally end-to-end receipts of the message
 *		Sent by same transport as for original message
 *	b) transport level : acknowledgement between facilitators, 
 *		signalling transport of the message, whether the message is
 *		an original message or the application-level ack.
 *
 * @param socket java.net.socket
 */
public void readAndDecipherMessage(java.net.Socket socket) {
	if (socket == null)
		return;
	try {
		MessagingProtocol origin = new MessagingProtocol(socket);
		try {
			DataInputStream dis = origin.getDataInputStream();
			ObjectInputStream ois = origin.getObjectInputStream();
			while (true) {
				byte tag = dis.readByte();
				if (tag == MessagingProtocol.MESSAGE) {
					Object obj = null;
					try {
						obj = ois.readObject();
						dis.readByte(); // read the EOF
					} catch (ClassNotFoundException ex) {
					}
					Message inMsg = null;
					if (obj instanceof Message) {
						inMsg = (Message) obj;
						//MobileCodeRecord mcRecord = residents.getMobileCodeRecord(inMsg.getTo());
						MobileCodeObject mcRecord = mcm.getMobileCodeHandle(inMsg.getTo());
						if (mcRecord != null) {
							synchronized (mcRecord) {
								if (inMsg.isRequest()) {
									// Message is a request, the original message
									// Destination netlet is local. Deliver message directly.
									deliverMessage(mcRecord, inMsg);
								} else {
									// Message is a response, an ack for a message sent previously
									sendMbx(ackMbx, inMsg.getTo(), inMsg);
								}
								origin.sendAck(MessagingProtocol.NO_ERROR);
							}
						} else {
							origin.sendAck(MessagingProtocol.DONT_EXIST);
						}
					} else {
						origin.sendAck(MessagingProtocol.ERROR);
					}
				}
			}
		} catch (IOException e) {
			origin.close();
		}
	} catch (java.io.IOException e) {
		mct.Console.error(null, getID() + " has bad connection on socket " + socket);
	}
}
/**
 * This method was created by a SmartGuide.
 * @return mct.mediator.Message
 * @param id java.lang.String
 */
public Message receive(String id) {
	return receive(id, 0);
}
/**
   * Receives a message, waiting if necessary
   * @param	id	Identifier of mobileCode that wishes to receive
   */

public Message receive(String id, long timer) {
	if (!available)
		return null;
//	MobileCodeRecord mcRecord = residents.getMobileCodeRecord(id);
	MobileCodeObject mcRecord = mcm.getMobileCodeHandle(id);
//	if ((mcRecord != null) && mcRecord.hasMessagingAccess) {
	if (mcRecord != null) {
		Message message = receiveMbx(inMbx, id, timer);
		if (message.isSync()) { 
			Message ack = new Message(message.getTo(), message.getFrom(), false);
			ack.addMessage("errorCode", "NO_ERROR");
//			MobileCodeRecord originator = residents.getMobileCodeRecord(message.getFrom());
			MobileCodeObject originator = mcm.getMobileCodeHandle(message.getFrom());
			if (originator != null) {
				// Originator of message is local. Deliver acknowledgement directly.
				sendMbx(ackMbx, ack.getTo(), ack);
			} else {
				// Originator of message is remote. Forward the acknowledgement via mediator
				if (!forwardMessage(ack)) {
					mct.Console.error(null, getID() + " received remote message from "+message.getFrom()+"\nbut unable to send remote ack to "+message.getTo());
				}
			}
		}
		return message;
	}
	return null;
}
/**
 * This method was created by a SmartGuide.
 * @return mct.mediator.Message[]
 * @param id java.lang.String
 */
public Message[] receiveAll(String id) {
	return receiveAll(id, 0);
}
/**
   * Receives all pending messages
   */
public Message[] receiveAll(String id, long timer) {
	if (!available)
		return null;
	// MobileCodeRecord mcRecord = residents.getMobileCodeRecord(id);
	MobileCodeObject mcRecord = mcm.getMobileCodeHandle(id);
	if ((mcRecord != null) && mcRecord instanceof MessagingAccess) {
		Message[] messages = receiveAllMbx(inMbx, id, timer);
		for (int i = 0; i < messages.length; i++) {
			Message message = messages[i];
			if (message.isSync()) {
				Message ack = new Message(message.getTo(), message.getFrom(), false);
				ack.addMessage("errorCode", "NO_ERROR");
				//MobileCodeRecord originator = residents.getMobileCodeRecord(message.getFrom());
				MobileCodeObject originator = mcm.getMobileCodeHandle(message.getFrom());
				if (originator != null) {
					// Originator of message is local. Deliver acknowledgement directly.
					sendMbx(ackMbx, ack.getTo(), ack);
				} else {
					// Originator of message is remote. Forward the acknowledgement via mediator
					if (!forwardMessage(ack)) {
						mct.Console.error(null, getID() + " received remote message, but unable to send remote ack");
					}
				}
			}
		}
		return messages;
	}
	return null;
}
/**
   * Retrieves a message from the dest's mailbox
   * NOTE : Default visibility is needed for FacilitatorNotifier
   * @param	mbxes	Table of all mailboxs, indexed by messengers' names
   * @param	dest	Name of receiver
   * @param time long
   * @return Message []
   */
Message[] receiveAllMbx(Hashtable mbxes, Object dest, long timer) {
	MailboxInterface mBox;
	if (!mbxes.containsKey(dest)) {
		if (mbxes == inMbx) {
			MobileCodeObject mCode = mcm.getMobileCodeHandle((String) dest);
			mBox = getMbx(mbxes, dest, mCode);
		} else
			mBox = new Mailbox();
		mbxes.put(dest, mBox);
	} else
		mBox = (MailboxInterface) mbxes.get(dest);
	return (Message[]) mBox.receiveAll(timer);
}
/**
   * Retrieves a message from the dest's mailbox
   * NOTE : Default visibility is needed for FacilitatorNotifier
   * @param	mbxes	Table of all mailboxs, indexed by messengers' names
   * @param	dest	Name of receiver
   * @param time long
   */
Message receiveMbx(Hashtable mbxes, Object dest, long timer) {
	MailboxInterface mBox;
	if (!mbxes.containsKey(dest)) {
		if (mbxes == inMbx) {
			MobileCodeObject mCode = mcm.getMobileCodeHandle((String) dest);
			mBox = getMbx(mbxes, dest, mCode);
		} else
			mBox = new Mailbox();
		mbxes.put(dest, mBox);
	} else
		mBox = (MailboxInterface) mbxes.get(dest);
	return (Message) mBox.receive(timer);
}
/**
   * The listening thread .... 
   * 
   * Receives messages forwarded from facilitators at other places.
   *
   * Also provides some initialization that cannot be done in the constructor
   * Could actually NOW be put in onInstantiate ()
   */
public void run() {

	// For each socket that we accept, we create a thread in order to run
	// the readAndDecipherMessage(java.net.Socket) method stored as part of the behaviour of 
	// the Facilitator (previously this work was performed by the FacilitatorReceiver).
	// Anything which has just a constructor and a run method can be replaced by
	// an instance of the MethodRunnerThread class.

	available = true;
	Object[] parameters = new Object[1];
	for (;;) {
		try {
			parameters[0] = listener.accept();
			new MethodRunnerThread(this, "readAndDecipherMessage", parameters, MCThreadNullExceptionHandler.getDefaultHandler());
		} catch (IOException e) {
			mct.Console.error(this, "Unable to accept incoming connection");
			e.printStackTrace();
		}
	}
}
/**
   * Sends the message to another netlet, whether local or remote.
   * Performs a synchronous send, with end-to-end acknowledge of 
   * destination's receipt.
   * @param	message		Message to be sent, including
   *				source and dest identifiers
   * @param	localOnly	If true, delivers message only
   *				if dest is on this same MCD
   *				If false, delivers message anywhere
   *				in the network, using the mediator
   *				to locate the dest
   * @returns	True when delivery is acknowledged, without error
   */

public boolean send(Message message, boolean localOnly) {
	return sendMessage(message, localOnly, false);
}
/**
   * Sends the message to another netlet, whether local or remote.
   * Performs an asynchronous send.
   * @param	message		Message to be sent, including
   *				source and dest identifiers
   * @param	localOnly	If true, delivers message only
   *				if dest is on this same MCD
   *				If false, delivers message anywhere
   *				in the network, using the mediator
   *				to locate the dest
   * @returns	True when delivery is acknowledged, without error
   */

public boolean sendAsync(Message message, boolean localOnly) {
	return sendMessage(message, localOnly, true);
}
/**
   * Puts the message into the dest's mailbox
   * NOTE : Default visibility is needed for FacilitatorNotifier
   * @param	mbxes	Table of mailboxes, indexed by messengers' names
   * @param	dest	Name of destination
   * @param	message Message to be put in mailbox
   */
void sendMbx(Hashtable mbxes, Object dest, Message message) {
	MailboxInterface mBox;
	if (!mbxes.containsKey(dest)) {
		if (mbxes == inMbx) {
			MobileCodeObject mCode = mcm.getMobileCodeHandle((String) dest);
			mBox = getMbx(mbxes, dest, mCode);
		} else
			mBox = new Mailbox();
		mbxes.put(dest, mBox);
	} else
		mBox = (MailboxInterface) mbxes.get(dest);
	mBox.send(message);
}
/**
   * Sends the message to another netlet, whether local or remote.
   * Performs an asynchronous send.
   * @param	message		Message to be sent, including
   *				source and dest identifiers
   * @param	localOnly	If true, delivers message only
   *				if dest is on this same MCD
   *				If false, delivers message anywhere
   *				in the network, using the mediator
   *				to locate the dest
   * @param async If true, the message is not acknowledged when read
   * 
   * @returns	True when delivery is acknowledged, without error
   */

public boolean sendMessage(Message message, boolean localOnly, boolean async) {
	if (!available)
		return false;
	String dest = message.getTo();
	if (dest.equalsIgnoreCase("ALL")) {
		System.err.println("Broadcast of messages to ALL is not yet handled");
		// Get an enumeration of all local netlets and send to each one
		// Could call this method recursively (for one-by-one messages)
		// or write another message that sends all then waits all
		// BUT !!! When you broadcast to ALL remotely, do you wait
		// for ACKS From ALL ????? I think not!  Therefore ALL
		// seems better handled by sendAsync ()
		return false;
	}
	//MobileCodeRecord mcRecord = residents.getMobileCodeRecord(message.getTo());
	MobileCodeObject mcRecord = mcm.getMobileCodeHandle(message.getTo());
	if (async)
		message.beAsync();
	else
		message.beSync();
	if (mcRecord != null) {
		// Destination netlet is local. Deliver message directly.
		if (!deliverMessage(mcRecord, message)) {
			return false; // Fails if netlet does not have MessagingAccess
		}
	} else
		if (!localOnly) {
			// Destination is remote, and message forwarding is turned on. 
			if (!forwardMessage(message)) {
				return false; // Fails if mediator is down OR if netlet not registered in mediator
			}
		} else {
			// Destination is remote, but message forwarding is turned off;
			return false;
		}
	return true;
}
public void setMediatorIP(InetAddress IPaddress) {
	mct.Console.warning(null,getID()+" Not allowed to set Mediator IP");
}
public void setMediatorPort(int port) {
	mct.Console.warning(null,getID()+" Not allowed to set Mediator port");
}
/** 
   * Sets the port on which the Facilitator is listening for messages
   * forwarded from other places
   * @param	port	New port for listening for incoming messages
   */
public void setPort(int port) {
	this.facPort = port;
}
/**
 * This method was created by Tony White.
 * If the mailbox already exists we must ensure that
 * the correct object is set up as that receiving
 * callbacks. This seems bizarre modelling and should
 * be revisited when time permits.
 * @return boolean
 * @param id java.lang.String
 */
public boolean startMbx(String id) {
	if (inMbx.contains(id)) {
		ThreadedMailbox mbox = (ThreadedMailbox) inMbx.get(id);
		MobileCodeObject mco = mcm.getMobileCodeHandle(id);
		if (mco != null) {
			mbox.setReceiver(new Receiver(mco, "onMessage", new Class[0]));
			return true;
		}
	}
	return false;
}
/**
 * This method was created by Tony White.
 * @return boolean
 * @param id String
 * @param force boolean
 */
public boolean stopMbx(String id, boolean force) {
	if (inMbx.contains(id)) {
		ThreadedMailbox mbox = (ThreadedMailbox)inMbx.get(id);
		if (mbox.stopMbx() || force) {
			inMbx.remove(id);
			ackMbx.remove(id);
		}
		else // Ensure that the mobile code does not see further callbacks
			mbox.setReceiver(null);
		return true;
	}
	return false;
}
/**
   * Registers this faciliator to a regional mediator to enable remote messaging
   * In this version, 
   *
   * @param listener class name for listener
   * @param listenerProperties file name for properties of listener
   * @param	host	Address of mediator's host
   * @param 	port	Listening port of mediator
   */

protected RemoteMCDirectory subscribeToMediator(String listener, String listenerProperties, InetAddress host, int port) {

	// First, load up the mediator's remote listener, so that local netlets
	// are tracked by the mediator

	// NOTE : Calling instantiate ASSUMES THAT THE CLASS CAN BE LOADED FROM THE CLASSPATH
	// IF YOU WANT TO BE ABLE TO LOAD FROM URL, THEN YOU WILL HAVE TO FIRST LOAD THE JAR !

	Properties props = new Properties();
	if (listenerProperties != null) {
		try {
			FileInputStream fis = new FileInputStream(listenerProperties);
			props.load(fis);
			fis.close();
		} catch (Exception ex) {
			mct.Console.error(this, "Could not load " + listenerProperties);
			return null;
		}
	}

	// Overwrite whatever is in the file, ensure consistency between the 
	// facilitator and listener
	props.put("listener.mediator.ip", host.getHostName());
	props.put("listener.mediator.name", mediatorName);
	if (!mcm.instantiateWithProperties(listener, props)) {
		// Disable remote messaging because without being tracked, 
		// we will not be able to receive our acknowledgements.
		return null;
	}
	MobileCodeObject handle;
	while ((handle = mcm.getMobileCodeHandle(listener)) == null) {
		Thread.yield();
	}
	if (handle instanceof RemoteMediatorListener) {
		while (!((RemoteMediatorListener) handle).isAvailable()) {
			try {
				Thread.sleep(150);
			} catch (java.lang.InterruptedException ex) {
			}
		}
	}
	else {
		mct.Console.error(this, "Remote mediator listener has incorrect interface");
	}
	// Second, get the handle of the remote directory, for future references
	RemoteMCDirectory rmcd = null;
	try {
		rmcd = (RemoteMCDirectory) Naming.lookup("//" + host.getHostName() + "/" + mediatorName);
		if (rmcd != null)
			while (!rmcd.isAvailable()) {
				try {
					Thread.sleep(150);
				} catch (java.lang.InterruptedException ex) {
				}
			}
	} catch (Exception e) {
		mct.Console.warning(this, e.getMessage());
	}
	return rmcd;
}
}