package mct.mediator;


/**
 * @(#) LocalMCDirectory.java	1.0 October-13-1997	Cheryl Schramm
 *
 *  The LocalMCDirectory is a netlet that is installed as
 *  an extension to the MobileCodeDaemon to maintain a
 *  directory of all mobile code resident on a single MobileCodeDaemon.  
 *  It is installed as an event listener, registering for 
 *  those events internal to the MobileCodeManager that pertain
 *  to the presence/movement of netlets within this MCD.
 *  As such, all changes to the MobileCodeDirectory occur by way
 *  of the listener methods; there are no public "add" methods.
 *  Other objects may however query this directory.
 *
 *  Why did I use the events I did ?
 *	
 *    * MobileCode was ADDED to this directory as soon
 *	as it was installed (MCEvent.INSTALLED).
 *	This event is valid for both netlets and vmcs
 *	(while the callback events only work for netlets)
 *    * MobileCode was REMOVED from this directory as soon
 *	as it was unable to receive messages (MigratableEvent.MIGRATED)
 */

import java.awt.*;
import java.util.*;
import mct.users.*;
import mct.admin.*;
public class LocalMCDirectory extends MobileCodeDirectory implements MCDInterface, MCEventListener, MigratableEventListener {
	private MobileCodeManager mcm;
	private MCTEventMulticaster eventDispatcher;
	private List list; // Used for debugging only

	String myLocation; // Location of "this" MCD, for convenience
	private Frame frame = null;
	private boolean _contentsOnly = true;
	private boolean available = false;


/**
  * 
  */
public LocalMCDirectory() {
	super();
	list = new List();
}
/**
  * @see mct.users.MCEventListener
  * When mobileCode is installed at a MobileCodeDaemon,
  * add a record, containing this mobileCode and its location,
  * in this LocalMCDirectory
  */
public void addMobileCode(MCEvent event) {
	MobileCodeObject mc = (MobileCodeObject) event.getSource();
	if (mc == null)
		return;
	synchronized (mc) {
		MobileCodeRecord mcr = new MobileCodeRecord(mc, this.myLocation);
		boolean okay = addMobileCodeRecord(mcr, mc.getEventNumber(), false);
		if (okay && frame != null) {
			if (_contentsOnly)
				list.addItem(mc.getID());
			else
				list.addItem(event.toString());
		}
	}
}
/**
  * @see mct.users.MCDExtension
  */
public MCTEventMulticaster getEventDispatcher() throws SecurityException {
	MobileCodeStub stub = getStub();
	if (stub != null) {
		return stub.getEventDispatcher();
	}
	return null;
}
/**
  * @see mct.users.MCDExtension
  */
public MobileCodeContainer getMobileCodeContainer() {
	MobileCodeStub stub = getStub();
	if (stub == null)
		return null;
	return stub.getMobileCodeContainer();
}
/**
  * @see mct.users.MCDExtension
  */
public Object getMobileCodeManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getMobileCodeManager();
	return null;
}
/**
  * @see mct.users.MCDExtension
  */
public StorageManager getStorageManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getStorageManager();
	return null;
}
/**
 * This method returns true when the directory is available
 * @return boolean
 */
public boolean isAvailable() {
	return available;
}
/**
  * @see mct.users.MigratableEventListener
  * When mobileCode is migrated,  remove this mobileCode's record 
  * from this LocalMCDirectory because it is no longer
  * available (in theory, at least)
  */
public void migrated(MigratableEvent event) {
	removeMobileCodeEntry((MobileCodeObject) event.getSource(), event.toString());
}
/**
  * @see 	MobileCodeObject
  *
  * Registers this object as a listener of events within the MobileCodeManager
  * 
  */
public boolean onInit(Properties properties) {
	if (properties != null) {
		if (properties.getProperty("directory.display.enable", "false").equals("true"))
			setupDisplay();
		_contentsOnly = properties.getProperty("directory.display.content", "summary").equals("summary");
		String title = properties.getProperty("directory.display.name");
		if (title != null)
			setTitle(title);
	}
	if ((mcm = (MobileCodeManager) getMobileCodeManager()) == null) {
		mct.Console.warning(this, "LocalMCDirectory cannot get MCM");
		return false;
	}

	// Register with the MobileCodeManager to receive
	// all events related to movement of netlets
	//    Callback, Migratable
	eventDispatcher = getEventDispatcher();
	if (eventDispatcher != null) {
		eventDispatcher.addMCEventListener((MCEventListener) this);
		eventDispatcher.addMigratableEventListener((MigratableEventListener) this);

		// For convenience ....
		myLocation = MobileCodeRecord.getLocation(mcm.getTcpPort(), mcm.getUdpPort(), mcm.getFacPort());
	} else {
		mct.Console.warning(this, "LocalMCDirectory is not active");
		return false;
	}
	available = true;
	return true;
}
/**
  * @see mct.users.MCEventListener
  * Do nothing when mobileCode is installed, to prevent
  * double-removal in conflict with migrated().
  * @param event MCEvent
  */
public void removeMobileCode(MCEvent event) {
	removeMobileCodeEntry((MobileCodeObject) event.getSource(), event.toString());
}
/**
  * @see mct.users.MCEventListener
  * Do nothing when mobileCode is installed, to prevent
  * double-removal in conflict with migrated().
  * @param mc mct.users.MobileCodeObject
  * @param event java.lang.String
  */
public void removeMobileCodeEntry(MobileCodeObject mc, String event) {
	synchronized (mc) {
		boolean okay = removeMobileCodeRecord(mc.getID(), mc.getEventNumber(), false);
		if (okay && frame != null) {
			if (_contentsOnly)
				list.remove(mc.getID());
			else
				list.addItem(event);
		}
	}
}
/**
 * This method was created by a SmartGuide.
 * @param title java.lang.String
 */
public void setTitle(String title) {
	if (frame != null)
		frame.setTitle(title);
}
/**
 * This method sets up a display for the local directory
 */
private void setupDisplay() {
	frame = new Frame("Local Directory");
	// list = new List (); Already set up in constructor
	frame.add(list);
	frame.pack();
	frame.validate();
	frame.show();
	return;
}
}