package mct.mediator;


/**
 * @(#) RemoteMediator.java	1.0 October-13-1997	Cheryl Schramm
 *
 *  The RemoteMediator is the remote version of a
 *  MobileCodeDirectory :
 *  The directory maintains a name-location listing
 *  of all mobileCode present in the MCD network,
 *  (within a region). 
 *  The directory is maintained by Remote listeners, 
 *  installed at each participating MCD, that monitor the movement
 *  of mobileCode at their locations and reflect these
 *  movements by calling the remote methods of this
 *  directory to add/remove/update entries.
 *
 *  Aside : The functionality of this directory is
 *      actually defined by the superclass, MobileCodeDirectory.
 *	This class is truly just a remote RMI interface to
 *	the MobileCodeDirectory
 */

import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.awt.*;
import java.util.*;
import java.io.*;
import mct.users.*;
import mct.admin.*;
public class RemoteMediator extends MobileCodeDirectory implements RemoteMCDirectory {
	private List eventList; // For debugging
	private List eventList2; // For debugging

	private static boolean _contentsOnly = true;
	private Frame frame; // For debugging
	private String mediatorName = "Mediator";
	private boolean available = false;








/**
   * Constructs the RemoteMediator as a remote server object,
   * exporting itself in the rmiregistry
   */
public RemoteMediator() {

	// Explicit call highlights that this class is simply
	// the remote RMI interface to the superclass
	super();
	try {
		UnicastRemoteObject.exportObject(this);
	} catch (RemoteException e) {
		e.printStackTrace();
	}
}
/**
   * The remote method for addMobileCodeRecord
   * @see MobileCodeDirectory
   */
public synchronized void addMobileCodeRecordR(MobileCodeRecord mcr, int event, boolean force) throws RemoteException {
	// Check whether this object is allowed to alter the directory
	boolean okay = super.addMobileCodeRecord(mcr, event, force);
	if (okay && frame != null) {
		if (_contentsOnly) {
			synchronized (eventList) {
				displayTable();
			}
		} else
			eventList.addItem("Adding " + mcr.id + " as " + mcr.toString());
	}
}
/**
 * This method was created by a SmartGuide.
 */
private void displayTable() {
	eventList.removeAll();
	eventList2.removeAll();
	MobileCodeRecord mcr = null;
	Enumeration records = getTable().elements();
	while (records.hasMoreElements()) {
		mcr = (MobileCodeRecord) records.nextElement();
		eventList.add(mcr.id);
		eventList2.add(mcr.location);
	}
	return;
}
/**
 * This method was created in VisualAge.
 */
public void finalize() {
	try {
		Naming.unbind(getMediatorName());
	} catch (Exception e) {
		mct.Console.error(null, "RemoteMediator unbind exception\nReason: " + e);
	}
}
/**
 * This method was created in VisualAge.
 * @return java.lang.String
 */
public String getMediatorName() {
	return mediatorName;
}
/**
   * The remote method for getMobileCodeRecord
   * @see MobileCodeDirectory
   */

public MobileCodeRecord getMobileCodeRecordR(String id) throws RemoteException {
	MobileCodeRecord temp = super.getMobileCodeRecord(id);
	if (!_contentsOnly) {
		if (temp == null) {
			eventList.addItem("Getting " + id + " failed");
		} else {
			eventList.addItem("Getting " + id + " is " + temp.toString());
		}
	}
	return super.getMobileCodeRecord(id);
}
/**
   * The remote method for hasMessagingAccess
   * @see MobileCodeDirectory
   */
public boolean hasMessagingAccessR(String id) throws RemoteException {
	return super.hasMessagingAccess(id);
}
/**
 * This method installs properties for the remote mediator.
 * A display can be created, for example.
 * @param propertiesFile java.lang.String
 */
private void installProperties(String propertiesFile) {
	Properties RMedProperties = new Properties();
	try {
		FileInputStream fis = new FileInputStream(propertiesFile);
		RMedProperties.load(fis);
		fis.close();
	} catch (IOException ex) {
		mct.Console.warning(this, "could not load mediator properties.\n" + ex);
		return;
	}

	// Now deal with the properties of interest: all others ignored
	mediatorName = RMedProperties.getProperty("directory.rmi.name", mediatorName);
	String name = null;
	name = RMedProperties.getProperty("directory.display.name", "Remote Mobile Code Directory");
	_contentsOnly = RMedProperties.getProperty("directory.display.content", "summary").equals("summary");
	String property;
	if (RMedProperties.getProperty("directory.display.enable", "false").equals("true"))
		setupDisplay(name);
	property = RMedProperties.getProperty("directory.security");
	if (property != null) {
		try {
			Class securityClass = Class.forName(property);
			System.setSecurityManager((SecurityManager) securityClass.newInstance());
		} catch (Exception ex) {
			mct.Console.error(this, "could not install security manager " + ex);
		}
	}
	return;
}
/**
 * This method returns true when the mediator is ready for business
 * @return boolean
 */
public boolean isAvailable() {
	return available;
}
/**
   * Startup program for the RemoteMediator :
   *	Creates the directory and exports the remote object
   *	to the rmiregistry
   */
public static void main(String args[]) {
	if (System.getSecurityManager() == null)
		System.setSecurityManager(new RMISecurityManager());
	RemoteMediator obj = null;
	try {
		obj = new RemoteMediator();
		if (args != null && args.length > 0)
			obj.installProperties(args[0]);
		Naming.rebind(obj.getMediatorName(), obj);
		obj.available = true;
		// Because no host is given in the URL, default = localhost
	} catch (Exception e) {
		mct.Console.error(null, obj.getMediatorName() + " RMI binding error");
		e.printStackTrace();
	}
}
/**
   * The remote method for mobileCodeElements
   * @see MobileCodeDirectory
   */
public Enumeration mobileCodeElementsR() throws RemoteException {
	return super.mobileCodeElements();
}
/**
   * The remote method for mobileCodeIDs
   * @see MobileCodeDirectory
   */
public Enumeration mobileCodeIDsR() throws RemoteException {
	return super.mobileCodeIDs();
}
/**
   * The remote method exported to remove a MobileCodeRecord from the directory
   * @see MobileCodeDirectory
   */

public synchronized void removeMobileCodeRecordR(String id, int event, boolean force) throws RemoteException {
	// Check whether this object is allowed to alter the directory
	MobileCodeRecord mcr = getMobileCodeRecord(id);
	boolean okay = super.removeMobileCodeRecord(id, event, force);
	if (okay && frame != null) {
		if (_contentsOnly) {
			synchronized (eventList) {
				displayTable();
			}
		} else
			eventList.addItem("Removing " + id + "(" + event + ")");
	}
}
/**
 * This method was created by a SmartGuide.
 * @param title java.lang.String
 */
public void setTitle(String title) {
	if (frame != null)
		frame.setTitle(title);
}
/**
 * This method was created by a SmartGuide.
 * @param name String
 */
private void setupDisplay(String name) {
	//* For debugging, visualizing only
	frame = new Frame(name);
	if (_contentsOnly)
		frame.setLayout(new GridLayout(1, 2));
	eventList = new List();
	frame.add(eventList);
	if (_contentsOnly) {
		eventList2 = new List();
		frame.add(eventList2);
	};
	frame.pack();
	frame.validate();
	frame.show();
	return;
}
}