package mct.mediator;


/**
 * @(#) RemoteMediatorListener.java	1.0 October-13-1997	Cheryl Schramm
 *
 *  The RemoteMediatorListener is a remote listener 
 *  for the RemoteMCDirectory.  It is installed as an
 *  extension to a MobileCodeDaemon, and it translates
 *  those events internal to the MobileCodeManager that pertain
 *  to the presence/movement of netlets within this MCD
 *  into add/remove operations on the RemoteMCDirectory.
 *
 *  In this version, the RemoteMCDirectory is a RMI server
 *  object, meaning that this listener calls the remote
 *  methods exported by the RemoteMCDirectory to effect
 *  changes to the directory
 *
 *  Maintaining the RemoteMCDirectory is sensitive to the
 *  ordering of events, particularly when events pertaining
 *  to a single mobileCode are distributed and thus may
 *  arrive in random order due to network delays.
 *  Consider : When a netlet migrates, several events occur :
 *	 On the original place : event = Migratable.MIGRATED
 *	 On the original place : event = MCEvent.DESTROYED
 *	 On the destination place : event = MCEvent.INSTALLED
 *  Since DESTROYED occurs after MIGRATED, INSTALLED may
 *	(and in fact, often does) arrive at the RemoteMCDirectory
 *	before DESTROYED.
 *
 *  For simplicity :
 *	1) INSTALLED adds this mobileCode to the RemoteMCDirectory
 *	   This event is also valid for both netlets and vmcs,
 *	   while the Callback events work only for netlets
 *	2) MIGRATED removes this mobileCode to the RemoteMCDirectory
 *	   Nothing is done when upon DESTROYED
 *  Consequentially, if a query to the RemoteMCDirectory fails, EITHER :
 *	* that mobileCode does not exist
 *	* that mobileCode is in transit, try again.
 *
 */

import java.awt.*;
import java.rmi.*;
import java.net.*;
import mct.users.*;
import mct.admin.*;
public class RemoteMediatorListener extends VirtualManagedComponent implements MCDInterface, MCEventListener, MigratableEventListener {
	private MobileCodeManager mcm;
	private MCTEventMulticaster eventDispatcher;
	private RemoteMCDirectory mediator;
	private boolean available = false;



/**
  * @see mct.users.MCEventListener
  * When mobileCode is installed at a MobileCodeDaemon,
  * add a record, containing this mobileCode and its location,
  * in the RemoteMCDirectory
  */
public void addMobileCode(MCEvent event) {
	MobileCodeObject mc = (MobileCodeObject) event.getSource();
	if (mc == null)
		return;
	String myLocation = MobileCodeRecord.getLocation(mcm.getTcpPort(), mcm.getUdpPort(), mcm.getFacPort());
	// NOTE : I tried having a class variable initialized in the run() startup
	//        but the ports were not YET initialized inside MobileCodeManager
	MobileCodeRecord mcr = new MobileCodeRecord(mc, myLocation);
	try {
		mediator.addMobileCodeRecordR(mcr, mc.getEventNumber(), false);
	} catch (NullPointerException ne) {
		mct.Console.error(null, " Mediator has not been set up yet. Can't log event");
	} catch (RemoteException re) {
		mct.Console.error(null, " Mediator is unable to log ADD change remotely");
	}
}
/**
  * @see mct.users.MCDExtension
  */
public MCTEventMulticaster getEventDispatcher() throws SecurityException {
	MobileCodeStub stub = getStub();
	if (stub != null) {
		return stub.getEventDispatcher();
	}
	return null;
}
/**
  * @see mct.users.MCDExtension
  */
public MobileCodeContainer getMobileCodeContainer() {
	MobileCodeStub stub = getStub();
	if (stub == null)
		return null;
	return stub.getMobileCodeContainer();
}
/**
  * @see mct.users.MCDExtension
  */
public Object getMobileCodeManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getMobileCodeManager();
	return null;
}
/**
  * @see mct.users.MCDExtension
  */
public StorageManager getStorageManager() {
	MobileCodeStub stub = getStub();
	if (stub != null)
		return stub.getStorageManager();
	return null;
}
/**
 * This method was created in VisualAge.
 * @return boolean
 */
public boolean isAvailable() {
	return available;
}
/**
  * @see mct.users.MigratableEventListener
  * When mobileCode is migrated,
  * remove this mobileCode's record from the RemoteMCDirectory
  * (because it is no longer valid)
  */
public void migrated(MigratableEvent e) {
	MobileCodeObject mc = (MobileCodeObject) e.getSource();
	try {
		mediator.removeMobileCodeRecordR(mc.getID(), mc.getEventNumber(), false);
	} catch (NullPointerException ne) {
		mct.Console.error(null, "Mediator has not been set up yet. Can't log event");
	} catch (RemoteException re) {
		mct.Console.error(null, "Mediator is unable to log REMOVE change remotely");
		re.printStackTrace();
	}
}
/**
  * @see	MobileCodeObject
  *
  * Registers this object as a listener of events within the MobileCodeManager
  * Locate handle of the RemoteMCDirectory to which these events will be forwarded
  */
public boolean onInit(java.util.Properties properties) {
	mediator = null;
	int tries = 0;
	String host;
	int maxTries = 5;
	long sleepTime = 1000;
	boolean okay = false;
	String mediatorName = "Mediator";
	mediatorName = properties.getProperty("listener.mediator.name", mediatorName);

	try {
		host = properties.getProperty("listener.mediator.ip", InetAddress.getLocalHost().getHostName());
	} catch (java.net.UnknownHostException hex) {
		host = "localhost";
	}
	String msg = "Could not lookup mediator on "+host+" with name: "+mediatorName;
	while (!okay && tries < maxTries) {
		try {
			tries++;
			mediator = (RemoteMCDirectory) Naming.lookup("//" + host + "/" + mediatorName);
			okay = true;
		} catch (Exception e) {
			try {
				Thread.sleep(sleepTime);
			} catch (InterruptedException iex) {
				mct.Console.warning(this, "waiting for mediator to come up");
				--tries;
			};
			msg = e.toString();
		}
	}

	// We're toast in terms of region-wide messaging
	// Could not find the Mediator, so give up.
	if (!okay) {
		mct.Console.warning(this, msg);
		return false;
	}
	if ((mcm = (MobileCodeManager) getMobileCodeManager()) == null) {
		mct.Console.error(this, " Cannot get Mobile Code Manager");
		return false;
	}
	eventDispatcher = getEventDispatcher();
	if (eventDispatcher != null) {
		eventDispatcher.addMCEventListener((MCEventListener) this);
		eventDispatcher.addMigratableEventListener((MigratableEventListener) this);
	} else {
		mct.Console.error(this, "Cannot get EventDispatcher");
		return false;
	}
	available = true;
	return true;
}
/**
  * @see mct.users.MCEventListener
  * Do nothing when mobileCode is de-installed, to prevent
  * double-removal in conflict with migrated()
  */
public void removeMobileCode(MCEvent event) {
	MobileCodeObject mc = (MobileCodeObject) event.getSource();
	try {
		mediator.removeMobileCodeRecordR(mc.getID(), mc.getEventNumber(), false);
	} catch (NullPointerException ne) {
		mct.Console.error(null, "Mediator has not been set up yet. Can't log event");
	} catch (RemoteException re) {
		mct.Console.error(null, "Mediator is unable to log REMOVE change remotely");
		re.printStackTrace();
	}
}
}