package mct.mediator;


/**
 * This type was created by Tony White.
 * It implements a threaded mailbox, where a callback is invoked whenever
 * a message is placed in the mailbox. Currently, if is only used for actual
 * messages -- not acknowledgements -- but this may change in the future.
 */
 import mct.users.Receiver;
public class ThreadedMailbox extends Thread implements MailboxInterface {
	private Mailbox mailBox;
	private Receiver receiver;
	private int count;
	private boolean processing;

/**
 * ThreadedMailbox constructor comment.
 * Set up the mailbox and the receiver;
 */
public ThreadedMailbox(Receiver receiver) {
	super("MBOX: "+receiver.toString());
	this.mailBox = new Mailbox();
	this.receiver = receiver;
	startMbx();
	setDaemon(true);
	start();
}
/**
 * This method was created by Tony White.
 * @return boolean
 */
public boolean hasMessages() {
	return mailBox.hasMessages();
}
/**
 * This method was created by Tony White.
 * @return java.lang.Object
 */
public Object peek() {
	return mailBox.peek();
}
/**
 * receive method comment. Access the
 * mailbox for messages
 */
public Object receive() {
	return receive(0);
}
/**
 * receive method comment. Access the
 * mailbox for messages
 */
public Object receive(long timer) {
	if (processing) {
		return mailBox.receive(timer);
	} else {
		System.out.println("TMBX: returning null");
		return null;
	}
}
/**
 * receiveAll method comment.
 */
public Object[] receiveAll() {
	return receiveAll(0);
}
/**
 * receiveAll method comment.
 @ param timer long (msecs)
 */
public Object[] receiveAll(long timer) {
	if (processing) {
		return mailBox.receiveAll(timer);
	}
	return null;
}
/**
 * This method was created by Tony White.
 * This is the work thread for the mailbox.
 * To start with we just suspend. The send method
 * resumes us in order that the callback can be invoked
 * for the message that has just been delivered. Messages
 * will be processed sequentially in this way. The "count"
 * variable ensures that the callback is invoked the appropriate
 * number of times.
 */
public void run() {
	suspend();
	Object[] obj = new Object[0];
	// For each message that arrives we need to invoke
	// the callback on that object. 
	while (processing) {
		while (count > 0 && receiver != null) {
			synchronized (receiver) {
				try {
					count--;
					receiver.invoke(obj);
				} catch (java.lang.IllegalAccessException ex) {
					mct.Console.warning(this, "Illegal Access Exception for " + receiver);
				} catch (java.lang.reflect.InvocationTargetException ex) {
					mct.Console.warning(this, "Invocation Target Exception for " + receiver);
				}	
			}
		}
		suspend();
	}
}
/**
 * Place a message into the mailbox.
 */
public synchronized void send(Object object) {
	count++;
	mailBox.send(object);
	resume();
}
/**
 * This method was created by Tony White.
 * We resume the thread as it might have messages
 * outstanding. We synchronize this method in order to
 * ensure that the callback completes before we set the 
 * receiver. Otherwise we could set the receiver to null while
 * performing a number of callbacks for several messages.
 *
 * @param receiver mct.users.Receiver
 */
public synchronized boolean setReceiver(Receiver receiver) {
	this.receiver = receiver;
	resume();
	return true;
}
/**
 * This method was created by Tony White.
 * Ensure that the system is processing messages.
 * @return boolean
 */
public synchronized boolean startMbx() {
	count = 0;
	return processing = true;
}
/**
 * This method was created by Tony White.
 * Ensure that the system is not processing messages.
 * This will only change the value of processing if
 * there are no messages in the mailbox. Otherwise we
 * would lose messages. We should probably have a lifetime
 * associated with a message.
 * @return boolean
 */
public synchronized boolean stopMbx() {
	if (mailBox.hasMessages())
		return processing;
	else {
		resume();
		return processing = false;
	}
}
}