package mct.net;


/**
 * NVTInputStream Class
 **/
/**************************************************************************
 *                    COPYRIGHT of MOBILE CODE TOOLKIT                    *
 *                                                                        *
 * This program is a free software and follows terms and conditions as    *
 * described on the GNU General Public License                            *
 * (see http://www.sce.carleton.ca/netmanage/gatot/gpl.html).             *
 * This software is freely distributed; however we recommend that         *
 * everybody should download a free copy directly from our site           *
 * http://www.sce.carleton.ca/netmanage/gatot.                            *
 * Therefore, we can keep a track who has got a copy and what for what    *
 * the purpose is. This will allow us to notify you when there are changes* 
 * or bugs fixed on the next release.                                     *                                   
 *                                                                        *
 * This program is a part of work in the Perpetuum Mobile Procura project.*
 * URL: http://www.sce.carleton.ca/netmanage/perpetuum.shtml.             *
 *                                                                        * 
 * Copyright 1997, Gatot Susilo                                           *
 * Department of Systems and Computer Engineering,                        *
 * Carleton University                                                    *
 *                                                                        *
 * Author: Gatot Susilo (susilo@sce.carleton.ca)                          *  
 * Date: October 5, 1997                                                  *
 * Contact Person:                                                        *
 *  Andrzej Bieszczad, Associate Professor, Systems & Computer Engineering*
 *  1125 Colonel By Drive, Carleton University, Ottawa, Canada K1S 5B6    *
 *  Email: andrzej@sce.carleton.ca    Tel:(613)520-5734  Fax:(613)520-5727*
 *  URL: http://www.sce.carleton.ca/faculty/bieszczad.html                *
 **************************************************************************
 *                                NO WARRANTY                             *
 *                                                                        *
 * BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY   *
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN*
 * OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES *
 * PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER       *
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED       *
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE*
 * ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH   *
 * YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL    *
 * NECESSARY SERVICING, REPAIR OR CORRECTION.                             *
 *                                                                        *    
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING  *
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR    *
 * REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR      *
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   *
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM         *
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED      *
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF *
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER OR*
 * OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.       *
 **************************************************************************/
import java.io.FilterInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.IOException;

/**
 * NVT stands for Network Virtual Terminal<br>
 * NVT classes implement minimal capability to establish
 * telnet-based application. It also provides for control
 * operations that support process interruption and the
 * discarding excessive output. These operation are signaled
 * by using the Interpret as Command (IAC) code.<p>
 *
 * Corresponding classes:
 * <ul>
 * <li>NVTInputStream<br>
 *     This class reads Input Stream and refuse negotiaon
 *     from other end. You may modify this, hence you
 *     can negotiate with other end.
 * <li>NVTOutputStream<br>
 *     This class sends Output Stream to other end and adds
 *     char CR and LF at the end of String.
 * <li>NVTPrinter<br>
 *     It writes to standard output and runs as a thread
 * </ul>
 */
public class NVTInputStream extends FilterInputStream {
	byte IAC = (byte) 0xff;
	byte DO = (byte) 0xfd;
	byte WILL = (byte) 0xfb;
	byte CR = 13;
	byte LF = 10;
	int WONT = 252;
	int DONT = 254;
	int BUFFER_SIZE = 1024;
	PrintWriter out;
	byte lineBuffer[] = new byte[BUFFER_SIZE];
	int numBytes = 0;



public NVTInputStream(InputStream inStream, PrintWriter outStream) {
	super(inStream);
	out = outStream;
}
/**
* It reads a byte from Input Stream and refuses negotiation
* from other end when it receives signal IAC.
* @return a next byte that represents as <em>int</em> range
*        from 0 to 255.
*/
public int read() throws IOException {
	boolean recIAC;
	int i;
	do {
		recIAC = false;
		i = in.read();
		if (i == -1)
			return i;
		byte b = (byte) i;
		if (b == IAC) {
			recIAC = true;
			int cmd = in.read();
			if (cmd == -1)
				return cmd;
			byte b2 = (byte) cmd;
			if (b2 == IAC)
				return 255;
			else
				if (b2 == DO) {
					int opt = in.read();
					if (opt == -1)
						return opt;
					out.write(255);
					out.write(WONT);
					out.write(opt);
					out.flush();
				} else
					if (b2 == WILL) {
						int opt = in.read();
						if (opt == -1)
							return opt;
						out.write(255);
						out.write(DONT);
						out.write(opt);
						out.flush();
					}
		}
	} while (recIAC);
	return i;
}
/**
* Just to make it convenience, you are allowed to read one
* line at the time other than one byte at the time.
* End of String is marked with char CR and LF
*/
public String readLine() throws IOException {
	numBytes = 0;
	boolean finished = false;
	do {
		int i = read();
		if (i == -1)
			return null;
		byte b = (byte) i;
		if (b == LF) {
			if (numBytes > 0) {
				if (lineBuffer[numBytes - 1] == 13)
					return new String(lineBuffer, 0, numBytes - 1);
			}
		}
		lineBuffer[numBytes] = b;
		++numBytes;
	} while (!finished);
	return null;
}
}